package minischeme

import java.io.PrintStream
import scala.collection.immutable.{Set,ListSet,Map,ListMap}
import TreeGenerator._

/**
 * Code generator for minischeme.
 *
 * @author Michel Schinz
 */
object Generator extends ((List[Tree],PrintStream) => Unit) {
  private val auxCodes = new scala.collection.mutable.ListBuffer[Code]()

  def apply(trees: List[Tree], stream: PrintStream): Unit = {
    // Attribute locations to global symbols
    var globalOffset: Int = 0
    trees foreach {
      case SList(SIdent("define") :: name :: _ :: Nil) =>
        name.symbol.location = GlobalLocation(globalOffset)
        globalOffset = globalOffset + 1
      case _ =>
    }

    // Allocate space for global variables
    val mainCode = new Code()
    mainCode.emit(LINT, Registers.GP, globalOffset)
    mainCode.emit(ALOC, Registers.GP, Registers.GP)

    // Allocate global "stack frame"
    val loadGlobalFrameSizeInstr = new InstructionRC(LINT, Registers.FP, 0)
    mainCode.emit(loadGlobalFrameSizeInstr)
    mainCode.emit(ALOC, Registers.FP, Registers.FP)

    // Generate program code
    trees foreach { gen(_, mainCode) }
    loadGlobalFrameSizeInstr.c = mainCode.getMaxFrameSize
    mainCode.emit(HALT)

    // Dump all code, starting with main one
    mainCode dumpTo stream
    auxCodes foreach (_ dumpTo stream)
  }

  private def gen(tree: Tree, code: Code): Unit = tree match {
    case SList(SIdent("define") :: name :: value :: Nil) =>
      code.addComment("begin " + name.symbol.name)
      code withFreshRegister { tmpReg =>
        genLoad(value, tmpReg, code)
        name.symbol.location match {
          case GlobalLocation(o) =>
            code.emit(STOR, tmpReg, Registers.GP, o)
        }
      }
      code.addComment("end " + name.symbol.name)

    case _ =>
      code withFreshRegister { tmpReg => genLoad(tree, tmpReg, code) }
  }

  private val binPrims =
    ListMap(("+", ADD), ("-", SUB), ("*", MUL), ("/", DIV), ("%", MOD),
            ("=", ISEQ), ("<", ISLT), ("<=", ISLE))

  private def genLoad(tree: Tree, targetReg: Register, code: Code): Unit =
  tree match {
    case SList(SIdent("lambda") :: SList(args) :: body) =>
      // Code for lambda: load label
      val label = Label.newLabel("lambda")
      code.emit(LINT, targetReg, label)

      // Code for lambda body
      val lCode = new Code()
      lCode.emit(label)
      for (argReg <- 1 to args.length)
        lCode.allocateRegister(Registers(argReg))
      // allocate and link frame
      var loadFrameSizeInstr: InstructionRC = null
      lCode withFreshRegister { tmpReg =>
        loadFrameSizeInstr = new InstructionRC(LINT, tmpReg, 0)
        lCode.emit(loadFrameSizeInstr)
        lCode.emit(ALOC, tmpReg, tmpReg)
        lCode.emit(STOR, Registers.FP, tmpReg, Registers.R0)
        lCode.incFrameSize(1)
        lCode.emit(ADD, Registers.FP, tmpReg, Registers.R0)
      }
      // save all parameters (including LK)
      lCode.emitPush(Registers.LK)
      for ((reg, arg) <- List.range(1, args.length + 1) zip args) {
        arg.symbol.location = FrameLocation(lCode.getFrameSize)
        lCode.emitPush(Registers(reg))
        lCode.freeRegister(Registers(reg))
      }
      val retValReg = lCode.allocateRegister(Registers(1))
      for (statement <- body)
        genLoad(statement, retValReg, lCode)
      lCode.decFrameSize(args.length)
      lCode.emitPop(Registers.LK)
      lCode.emitPop(Registers.FP)
      assert(lCode.getFrameSize == 0, "non-zero frame size")
      lCode.emit(JMPZ, Registers.LK, Registers.R0) // return
      lCode.freeRegister(retValReg)
      loadFrameSizeInstr.c = lCode.getMaxFrameSize
      auxCodes += lCode

    case SList(SIdent("if") :: cond :: thenp :: elsep :: Nil) =>
      genLoad(cond, targetReg, code)
      val elseLabel = Label.newLabel("else")
      code.emitJump(elseLabel, targetReg)
      genLoad(thenp, targetReg, code)
      val afterLabel = Label.newLabel
      code.emitJump(afterLabel, Registers.R0)
      code.emit(elseLabel)
      genLoad(elsep, targetReg, code)
      code.emit(afterLabel)

    case SList(SIdent("let") :: SList(bindings) :: body) =>
      for (SList(ident :: value :: Nil) <- bindings) {
        genLoad(value, targetReg, code)
        ident.symbol.location = FrameLocation(code.getFrameSize)
        code.emitPush(targetReg)
      }
      for (statement <- body)
        genLoad(statement, targetReg, code)
      code.decFrameSize(bindings.length)

    case SNum(value) =>
      code.emit(LINT, targetReg, value)

    case SList(SIdent(ident) :: e1 :: e2 :: Nil) if (binPrims contains ident) =>
      genLoad(e1, targetReg, code)
      code withFreshRegister { e2Reg =>
        genLoad(e2, e2Reg, code)
        code.emit(binPrims(ident), targetReg, targetReg, e2Reg)
      }

    case SList(SIdent("vector") :: elems) =>
      code.emit(LINT, targetReg, elems.length)
      code.emit(ALOC, targetReg, targetReg)
      code withFreshRegister { eReg =>
        var offset = 0
        for (e <- elems) {
          genLoad(e, eReg, code)
          code.emit(STOR, eReg, targetReg, offset)
          offset = offset + 1
        }
      }

    case SList(SIdent("vector-ref") :: vect :: SNum(0) :: Nil) =>
      genLoad(vect, targetReg, code)
      code.emit(LOAD, targetReg, targetReg, Registers.R0)

    case SList(SIdent("vector-ref") :: vect :: index :: Nil) =>
      genLoad(vect, targetReg, code)
      code withFreshRegister { indexReg =>
        genLoad(index, indexReg, code)
        code.emit(LOAD, targetReg, targetReg, indexReg)
      }

    case SList(SIdent("vector-set!") :: vect :: index :: value :: Nil) =>
      genLoad(vect, targetReg, code)
      code withFreshRegister { indexReg =>
        genLoad(index, indexReg, code)
        code.withFreshRegister { valueReg =>
          genLoad(value, valueReg, code)
          code.emit(STOR, valueReg, targetReg, indexReg)
        }
      }

    case SList(SIdent("print-char") :: x :: Nil) =>
      genLoad(x, targetReg, code)
      code.emit(PCHR, targetReg)
    case SList(SIdent("read-char") :: Nil) =>
      code.emit(RCHR, targetReg)

    case SList(function :: args) =>
      // save registers
      code.freeRegister(targetReg)
      val registersToSave = code.allocatedRegisters
      for (r <- registersToSave) {
        code.emitPush(r)
        code.freeRegister(r)
      }
      val funReg = code.allocateRegister(Registers(28))
      genLoad(function, funReg, code)
      for (i <- 1 to args.length) {
        code.allocateRegister(Registers(i))
        genLoad(args(i - 1), Registers(i), code)
      }
      val retLabel = Label.newLabel("ret")
      code.emit(LINT, Registers.LK, retLabel)
      code.emit(JMPZ, funReg, Registers.R0)
      code.emit(retLabel)
      code.freeRegister(funReg)
      for (i <- 1 to args.length)
        code.freeRegister(Registers(i))
      val retValReg = Registers(1)
      code.allocateRegister(targetReg)
      if (targetReg != retValReg)
        code.emit(ADD, targetReg, retValReg, Registers.R0)
      // restore registers
      for (r <- registersToSave.reverse) {
        code.allocateRegister(r)
        code.emitPop(r)
      }

    case SIdent(name) =>
      tree.symbol.location match {
        case FrameLocation(index) =>
          code.emit(LOAD, targetReg, Registers.FP, index)
        case GlobalLocation(index) =>
          code.emit(LOAD, targetReg, Registers.GP, index)
      }

    case _ =>
      Report.error(tree.pos, "invalid tree: " + tree)
  }
}
