package minischeme

/**
 * Interpreter for minischeme
 *
 * @author Iulian Dragos, Michel Schinz
 */
object Interpreter extends (List[Tree] => Unit) {
  // Values
  abstract class Value {
    override def toString(): String = this match {
      case Lambda(_) => "<lambda>"
      case Vector(_) => "<vector>"
      case Integer(v) => v.toString()
    }
  }
  case class Lambda(body: List[Value] => Value) extends Value
  case class Vector(values: Array[Value]) extends Value
  case class Integer(v: Int) extends Value

  // Error handling
  case class InterpreterError() extends Exception()

  def printStackEntry(pos: Position): Unit =
    Console.println("  at line " + pos.line + " in file " + pos.file)

  def error(pos: Option[Position], msg: String): Nothing = {
    Console.println("Error: " + msg)
    pos foreach printStackEntry
    throw new InterpreterError()
  }
  def error(pos: Position, msg: String): Nothing = error(Some(pos), msg)
  def error(msg: String): Nothing = error(None, msg)

  // Environment
  type Env = PartialFunction[String, Value]
  def makeEnv(bdgs: List[(String,Value)]): Env =
    (new scala.collection.immutable.ListMap()) ++ bdgs

  val globalEnv = new scala.collection.mutable.HashMap[String, Value]

  def apply(trees: List[Tree]): Unit = try {
    trees foreach {
      case SList(List(SIdent("define"), SIdent(name), _)) =>
        globalEnv += name -> Integer(0)
      case _ =>
    }
    trees foreach {
      case SList(SIdent("define") :: SIdent(name) :: body :: Nil) =>
        globalEnv += name -> eval(body, globalEnv)
      case tree =>
        eval(tree, globalEnv)
    }
  } catch {
    case InterpreterError() => exit(1)
  }

  def evalN(trees: List[Tree], env: Env): Value =
    (trees map { t: Tree => eval(t, env) }).last

  def eval(tree: Tree, env: Env): Value = {
    def t_error(msg: String): Nothing = error(Some(tree.pos), msg)

    tree match {
      case SList(SIdent("lambda") :: SList(params) :: body) =>
        val paramNames = params map {
          case SIdent(name) => name
          case other => error(other.pos, "malformed parameter")
        }
	Lambda({ args: List[Value] =>
	  if (args.length == params.length)
	    evalN(body, makeEnv(paramNames zip args) orElse env)
	  else
	    error("wrong number of arguments (expected " + params.length
                  + ", got " + args.length + ")")
        })
      case SList(SIdent("lambda") :: _) =>
        t_error("malformed lambda")

      case SList(SIdent("let") :: SList(bdgs) :: body) =>
        val evalBdgs = bdgs map {
          case SList(SIdent(ident) :: value :: Nil) => (ident, eval(value, env))
          case other => error(other.pos, "malformed binding")
        }
	evalN(body, makeEnv(evalBdgs) orElse env)
      case SList(SIdent("let") :: _) =>
        t_error("malformed let")

      case SList(SIdent("if") :: cond :: thenp :: elsep :: Nil) =>
	eval(cond, env) match {
	  case Integer(0) => eval(elsep, env)
	  case _ => eval(thenp, env)
	}
      case SList(SIdent("if") :: _) =>
        t_error("malformed if")

      case SList(SIdent("read-char") :: Nil) =>
        Integer(System.in.read())

      case SList(SIdent("print-char") :: c :: Nil) =>
        eval(c, env) match {
          case Integer(i) =>
            System.out.print(i.asInstanceOf[Char])
            System.out.flush()
            Integer(0)
	  case other =>
            t_error("integer expected, got " + other)
	}

      case SList(SIdent("vector") :: args) =>
        Vector((args map { a: Tree => eval(a, env) }).toArray)

      case SList(SIdent("vector-ref") :: vect :: index :: Nil) =>
        (eval(vect, env), eval(index, env)) match {
          case (Vector(contents), Integer(i)) =>
            if (0 <= i && i < contents.length)
              contents(i)
            else
              t_error("index " + i + " not in [0;" + contents.length + ")")
          case (other, Integer(_)) =>
            t_error("vector expected for vector-ref, got " + other)
          case (_, other) =>
            t_error("integer expected for vector-ref index, got " + other)
        }

      case SList(SIdent("vector-set!") :: vect :: index :: value :: Nil) =>
        (eval(vect, env), eval(index, env), eval(value, env)) match {
          case (Vector(contents), Integer(i), v) =>
            if (0 <= i && i < contents.length) {
              contents(i) = v
              Integer(0)
            } else
              t_error("index " + i + " not in [0;" + contents.length + ")")
          case (other, Integer(_), _) =>
            t_error("vector expected for vector-set!, got " + other)
          case (_, other, _) =>
            t_error("integer expected for vector-set! index, got " + other)
        }

      case SList((f @ SIdent(name)) :: e1 :: e2 :: Nil) if Primitives.is(f) =>
        (name, eval(e1, env), eval(e2, env)) match {
          case ("+", Integer(v1), Integer(v2)) => Integer(v1 + v2)
          case ("-", Integer(v1), Integer(v2)) => Integer(v1 - v2)
          case ("*", Integer(v1), Integer(v2)) => Integer(v1 * v2)
          case ("/", Integer(v1), Integer(v2)) => Integer(v1 / v2)
          case ("%", Integer(v1), Integer(v2)) => Integer(v1 % v2)

          case ("=", Integer(v1), Integer(v2)) =>
            Integer(if (v1 == v2) 1 else 0)
          case ("=", Vector(v1), Vector(v2)) =>
            Integer(if (v1 eq v2) 1 else 0)
          case ("<", Integer(v1), Integer(v2)) =>
            Integer(if (v1 < v2) 1 else 0)
          case ("<=", Integer(v1), Integer(v2)) =>
            Integer(if (v1 <= v2) 1 else 0)

          case (name, v1, v2) =>
            t_error("invalid arguments for " + name + ": " + v1 + " " + v2)
        }
      case SList((f @ SIdent(name)) :: _) if Primitives.is(f) =>
        t_error("malformed primitive use")

      case SList(fun :: args) =>
        eval(fun, env) match {
	  case Lambda(body) =>
            try {
              body(args map { a: Tree => eval(a, env) })
            } catch {
              case e @ InterpreterError() =>
                printStackEntry(tree.pos)
                throw e
            }
	  case v =>
            t_error("expected function, got " + v)
	}

      case SIdent(name) =>
        if (env isDefinedAt name)
          env(name)
        else
          t_error("unknown identifier " + name)

      case SNum(value)  =>
        Integer(value)

      case _ =>
        t_error("unexpected tree: " + tree)
    }
  }
}
