package minischeme

import scala.collection.immutable.Map

/**
 * Name analysis for minischeme.
 *
 * @author Michel Schinz
 */
object NameAnalyzer extends ((Map[String,Symbol],List[Tree]) => Unit) {
  private type Env = Map[String,Symbol]

  def apply(env: Env, trees: List[Tree]): Unit = {
    val globalEnv = collectGlobalEnv(env, trees)
    for (tree <- trees) analyze(globalEnv, tree)
  }

  private def collectGlobalEnv(init: Env, trees: List[Tree]): Env =
    trees.foldLeft(init) { (env, tree) =>
      tree match {
        case SList(SIdent("define") :: SIdent(name) :: _) =>
          env + name -> new Symbol(name, GlobalKind)
        case _ =>
          env
      }
    }

  private def analyze(env: Env, tree: Tree): Unit = tree match {
    case SList(SIdent("define") :: (t @ SIdent(name)) :: value :: Nil) =>
      t.symbol = env(name)
      analyze(env, value)

    case SList(SIdent("lambda") :: SList(parameters) :: body) =>
      val symbols = for (t @ SIdent(name) <- parameters) yield {
        t.symbol = new Symbol(name, LocalKind)
        (name, t.symbol)
      }
      checkDefs(tree.pos, symbols map (_._1))
      val augmentedEnv = env ++ symbols
      for (statement <- body)
        analyze(augmentedEnv, statement)

    case SList(SIdent("let") :: SList(bdgs) :: body) =>
      val symbols = for (SList((t @ SIdent(name)) :: rhs :: Nil) <- bdgs) yield{
        t.symbol = new Symbol(name, LocalKind)
        analyze(env, rhs)
        (name, t.symbol)
      }
      checkDefs(tree.pos, symbols map (_._1))
      val augmentedEnv = env ++ symbols
      for (statement <- body)
        analyze(augmentedEnv, statement)

    case SList(SIdent("if") :: cond :: thenp :: elsep :: Nil) =>
      analyze(env, cond)
      analyze(env, thenp)
      analyze(env, elsep)

    case SList(elements) =>
      for (element <- elements)
        analyze(env, element)

    case SIdent(name) =>
      if (env contains name)
        tree.symbol = env(name)
      else
        Report.error(tree.pos, "unknown identifier " + name)

    case SNum(_) =>
  }

  private def checkDefs(pos: Position, names: List[String]): Unit = {
    val defined = new scala.collection.mutable.HashSet[String]()
    for (name <- names) {
      if (defined(name))
        Report.error(pos, "multiple definition of " + name)
      defined += name
    }
  }
}
